import os, sys, shutil, re, codecs, subprocess, pwd, argparse
import yaml
from jinja2 import Environment, FileSystemLoader

import watchdog, time
from watchdog.observers import Observer
from watchdog.events import LoggingEventHandler
from threading import Thread

MEDIA_EXTS = ['.png',
			  '.gif',
			  '.jpg',
			  '.ico',
			  '.txt',
			  '.svg',
			  '.css',
			  '.eot',
			  '.ttf',
			  '.woff',
			  '.swf',
			  '.zip',]

def startswith(val, pattern):
	# TODO: This is weird
	if val:
		return val.startswith(pattern)
	else:
		return ""

class PureStatic:
	
	def __init__(self, configFile, sourceFolder, destinationFolder):
		# Record important file system locations
		self.sourceFolder = sourceFolder
		self.destinationFolder = destinationFolder

		# Load the configuration file
		with open(configFile) as f:
			# Replace tabs with spaces to allow tabs in YAML file
			configText = f.read().replace("\t", "    ")
		self.configData = yaml.load(configText)

		# Initialize Jinja
		self.templateEnv = Environment(loader=FileSystemLoader(sourceFolder))
		self.templateEnv.filters['startswith'] = startswith
		
	def delete_file(self, absFilePath):
		# TODO: This is all redundant with `build_file()`
		
		# Figure out the file's relative path
		relativeFile = os.path.relpath(absFilePath, self.sourceFolder)
		
		# Split path components
		sourceDir, sourceBasename = os.path.split(absFilePath)
		ext = os.path.splitext(sourceBasename)[1]
		
		# Delete the destination file
		destinationFile = os.path.join(self.destinationFolder, relativeFile)
		try:
			os.remove(destinationFile)
		except OSError:
			pass
	
	def build_file(self, absFilePath):
		# TODO: Record dependencies (CSS and templates)
		# TODO: Ignore `config` directory
		
		# Figure out the file's relative path
		relativeFile = os.path.relpath(absFilePath, self.sourceFolder)
		
		# Split path components
		sourceDir, sourceBasename = os.path.split(absFilePath)
		ext = os.path.splitext(sourceBasename)[1]

		if ext in ['.nginx', '.html']:
			destinationFile = os.path.join(self.destinationFolder,
										   relativeFile)
			
			print("Updating: {0}".format(absFilePath))
			
			# Check if any subdirs have to be created
			destinationSubdir = os.path.dirname(destinationFile)
			if not os.path.isdir(destinationSubdir):
				os.makedirs(destinationSubdir)
			
			# Load the file as a Jinja template
			template = self.templateEnv.get_template(relativeFile)
			
			# Render the Jinja template
			page = template.render(self.configData)
			
			with codecs.open(destinationFile, 'wb', 'utf-8') as f:
				f.write(page)
		
			if ext == '.nginx': # TODO: .nginx ext is now special
				print("Reloading nginx config files")
				subprocess.call(['sudo', 'nginx', '-s', 'reload'])
		
		# TODO: process SCSS files, coffeescript, anything else?
		elif ext in MEDIA_EXTS:
			# Copy media files over without modification
			destinationFile = os.path.join(self.destinationFolder, relativeFile)
			
			# Check if we actually need to do anything (TODO: Bad?)
			if os.path.isfile(destinationFile):
				sourceTime = os.path.getmtime(absFilePath)
				destinationTime = os.path.getmtime(destinationFile)
				if sourceTime <= destinationTime:
					return
			
			print("Updating: {0}".format(absFilePath))
			
			# Check if any subdirs have to be created
			destinationSubdir = os.path.dirname(destinationFile)
			if not os.path.isdir(destinationSubdir):
				os.makedirs(destinationSubdir)

			shutil.copyfile(absFilePath, destinationFile)
	
	def start_watching(self):
		# Define a watcher
		class PureStaticWatcher(watchdog.events.FileSystemEventHandler):

			def __init__(self, pureStatic):
				self.pureStatic = pureStatic
			
			def on_created(self, event):
				if event.is_directory:
					return
				# Rebuild the file
				self.pureStatic.build_file(event.src_path)

			def on_modified(self, event):
				if event.is_directory:
					return
				# Rebuild the file
				self.pureStatic.build_file(event.src_path)

			def on_deleted(self, event):
				if event.is_directory:
					return
				# Delete the file in the destination directory
				self.pureStatic.delete_file(event.src_path)
		
		# Start watching files
		eventHandler = PureStaticWatcher(self)
		self.observer = Observer()
		self.observer.schedule(eventHandler, self.sourceFolder, recursive=True)
		self.observer.start()

	def build(self):
		for root, dirs, files in os.walk(self.sourceFolder):
			for filename in files:
				self.build_file(os.path.join(root, filename))
	
	def watch(self):
		self.start_watching()
		try:
			while True:
				time.sleep(1)
		except KeyboardInterrupt:
			self.observer.stop()

def get_main_nginx_config():
	pipe = subprocess.PIPE
	p = subprocess.Popen(['sudo', 'nginx', '-t'],
						 stdin=pipe,
						 stdout=pipe,
						 stderr=pipe)
	output = p.communicate("")[1]

	# Try a successful `nginx -t` test
	mainNginxConfig = None
	match = re.search("nginx: the configuration file (.+) syntax is ok", output)
	if match:
		mainNginxConfig = match.group(1)
	
	# Try a failed `nginx -t` test
	match = re.search("nginx: configuration file (.+) test failed", output)
	if match:
		mainNginxConfig = match.group(1)

	if not mainNginxConfig:
		# Maybe the `nginx -t` test failed
		print("ERROR: Could not locate the main nginx configuration file")
		sys.exit(1)
	
	return mainNginxConfig

def get_nginx_symlink_file(domain):
	mainNginxConfig = get_main_nginx_config()
	with open(mainNginxConfig) as f:
		config = f.read()
	match = re.search(r"^[^#]*include\s+(.*servers|.*sites-enabled)/\*;$",
					  config,
					  re.MULTILINE)
	nginxIncludeDir = match.group(1)
	if not os.path.isabs(nginxIncludeDir):
		# Relative links need to be relative to the main nginx.conf file
		nginxIncludeDir = os.path.join(os.path.dirname(mainNginxConfig), nginxIncludeDir)
	return os.path.join(nginxIncludeDir, domain)

def create_site(domain, isServer=False):
	# TODO: This could maybe be abstracted to put folders anywhere
	
	# Need `uid` for setting the correct owner when run as `sudo`
	user = os.getlogin()
	uid = pwd.getpwnam(user).pw_uid
	rootuid = os.getuid()
	
	# Make sure the project directory exists
	sitesDir = os.path.expanduser("/var/www")
	if not os.path.isdir(sitesDir):
		print("Creating {0}".format(sitesDir))
		os.makedirs(sitesDir)
		os.chown(sitesDir, uid, -1)
	
	# Make sure it has the correct permissions
	permissions = oct(os.stat(sitesDir).st_mode & 0777)
	if not permissions == "0755":
		print("Changing permissions of {0} to 755".format(sitesDir))
		os.chmod(sitesDir, 0755)
	
	# --- TODO: Above can be abstracted --

	# Add the domain to /etc/hosts
	if not isServer:
		# Only need to edit /etc/hosts if local repo
		with open("/etc/hosts", "a") as f:
			print("Adding {0} to /etc/hosts".format(domain))
			f.write("127.0.0.1 {0}\n".format(domain)) # TODO: Needs to be real IP?
	
	# Create the project/ directory
	projectDir = os.path.join(sitesDir, domain)
	if not os.path.isdir(projectDir):
		print("Creating {0}".format(projectDir))
		os.makedirs(projectDir)
		os.chown(projectDir, uid, -1)
	
	# Create the source/ directory
	sourceDir = os.path.join(projectDir, "source")
	if not os.path.isdir(sourceDir):
		print("Creating {0}".format(sourceDir))
		os.makedirs(sourceDir)
		os.chown(sourceDir, uid, -1)

	# Create a dummy config.nginx file
	nginxConfigFile = os.path.join(sourceDir, "config.nginx")
	with open(nginxConfigFile, 'w') as f:
		f.write("""server {
	listen       80;
	server_name  {{ serverName }};

	root   {{ projectRoot }}/site/public;
	
	access_log {{ projectRoot }}/logs/access.log;
	error_log {{ projectRoot }}/logs/error.log;

	rewrite /blah / permanent;

	location / {
		index  index.html;
	}

	#error_page  404              /404.html;

	# redirect server error pages to the static page /50x.html
	#
	error_page   500 502 503 504  /50x.html;
	location = /50x.html {
		root   html;
	}
}""")
	os.chown(nginxConfigFile, uid, -1)
	
	# Create the source/public/ directory
	publicDir = os.path.join(sourceDir, "public")
	if not os.path.isdir(publicDir):
		print("Creating {0}".format(publicDir))
		os.makedirs(publicDir)
		os.chown(publicDir, uid, -1)
	
	# Create a dummy index.html file
	indexFile = os.path.join(publicDir, "index.html")
	with open(indexFile, 'w') as f:
		f.write("""<!doctype>
<html>
<head>
	<title>{{ websiteName }}</title>
</head>
<body>
	<h1>Hello, World!</h1>
	<p>This template should get you started.</p>
</body>
</html>
""")
	os.chown(indexFile, uid, -1)
	
	# Create the source/config/ directory
	configDir = os.path.join(sourceDir, "config")
	if not os.path.isdir(configDir):
		print("Creating {0}".format(configDir))
		os.makedirs(configDir)
		os.chown(configDir, uid, -1)

	# Create a dummy config file
	configFile = os.path.join(configDir, domain)
	with open(configFile, 'w') as f:
		f.write("""websiteName: Example Website
serverName: {0}
projectRoot: {1}""".format(domain, projectDir))
	os.chown(configFile, uid, -1)
	
	# Create the site/ directory
	siteDir = os.path.join(projectDir, "site")
	if not os.path.isdir(siteDir):
		print("Creating {0}".format(siteDir))
		os.makedirs(siteDir)
		os.chown(siteDir, uid, -1)
	
	# Create the logs/ directory
	logsDir = os.path.join(projectDir, "logs")
	if not os.path.isdir(logsDir):
		print("Creating {0}".format(logsDir))
		os.makedirs(logsDir)
		# TODO: Leaving owner as root is ok?
	
	# Find the main nginx config file (test this on other systems and nginx
	# versions)
	mainNginxConfig = get_main_nginx_config()

	
	# Create a symbolic link in either `sites-enabled` or `server`
	# (Depends on OS/nginx install method)
	nginxLinkFile = os.path.join(projectDir, "site/config.nginx")
	nginxSymlinkFile = get_nginx_symlink_file(domain)
	print("Creating symbolic link from {0} to {1}".format(nginxSymlinkFile,
		  nginxLinkFile))
	compiledNginxConfigFile = os.path.join(siteDir, "config.nginx")
	os.symlink(compiledNginxConfigFile, nginxSymlinkFile)


	# TODO: Consistent path variables
	# TODO: Uninstalling is a PITA. Should fail gracefully and cleanup after
	# itself (i.e., delete /etc/hosts entry).
	# Build once so nginx config isn't pointing to a missing file
	pureStatic = PureStatic(configFile,
							sourceDir,
							siteDir)
	
	pureStatic.build()

	# The build() is run as root, which is not good. Change to user.
	# TODO: THIS IS UGLY AS HELL
	for root, dirs, files in os.walk(siteDir):  
		for folder in dirs:  
			os.chown(os.path.join(root, folder), uid, -1)
			os.chmod(os.path.join(root, folder), 0755)
		for f in files:
			os.chown(os.path.join(root, f), uid, -1)
			os.chmod(os.path.join(root, f), 0755)

	# -- GIT SETUP --

	# Change to a normal user so Git repo permissions are correct
 	os.setuid(uid) # CANT CHANGE BACK TO ROOT AFTER THIS

	# Create a .gitignore file
	gitIgnoreFile = os.path.join(projectDir, ".gitignore")
	with open(gitIgnoreFile, 'w') as f:
		f.write("""site
logs
""")
	os.chmod(gitIgnoreFile, 0755)

	# Initialize a Git repo
	os.chdir(projectDir) # TODO: THIS COULD BE BAD. Need for Git calls.
	subprocess.call(['git', 'init', '.'])
	
	# Set up for pushes if necessary
	if isServer:
		print("SETTING UP SERVER")
		subprocess.call(['git', 'config', 'receive.denyCurrentBranch', 'ignore'])
		
		hookFile = os.path.join(projectDir, '.git/hooks/post-receive') # TODO
		with open(hookFile, 'w') as f:
			f.write("""#!/bin/bash

branchToDeploy=master

echo "Deploying $branchToDeploy branch"
git --work-tree=.. checkout $branchToDeploy
git --work-tree=.. checkout -f

# Build the site
sudo purestatic build {0} # TODO
""".format(domain))
	
		# Set owner and execute permissions on hook file
		os.chmod(hookFile, 0755)
	
	
def delete_site(domain, isServer=False):
	# Delete the project director in ~/www
	#sitesDir = os.path.expanduser("~/www")
	sitesDir = os.path.expanduser("/var/www")
	projectDir = os.path.join(sitesDir, domain)
	if os.path.isdir(projectDir):
		shutil.rmtree(projectDir)
	
	# Remove the domain from /etc/hosts
 
	if not isServer:
		with open("/etc/hosts") as f:
			hosts = f.read()
		hosts = hosts.replace("127.0.0.1 {0}\n".format(domain), "")
		with open("/etc/hosts", "w") as f:
			print("Removing {0} from /etc/hosts".format(domain))
			f.write(hosts)

	
	# Remove the symbolic link from the main nginx configuration
	nginxLinkFile = os.path.join(projectDir, "site/config.nginx")
	nginxSymlinkFile = get_nginx_symlink_file(domain)
	print("Removing symbolic link from {0} to {1}".format(nginxSymlinkFile,
		  nginxLinkFile))
	try:
		os.unlink(nginxSymlinkFile)
	except OSError:
		print("No symbolic link found")
	
	# Reload the nginx configuration
	print("Reloading nginx config files")
	subprocess.call(['sudo', 'nginx', '-s', 'reload'])



def main():
	description = "Pure Static Website Generator"
	parser = argparse.ArgumentParser(description=description)
	parser.add_argument('subcommand',
						choices=['create', 'delete', 'build', 'watch'],
						help='What do you want to do?')
	parser.add_argument('domain',
						help='The domain to operate on (e.g., example.com)')
	parser.add_argument('--server', dest='server', action='store_true',
						help='Whether you want to allow Git pushes to this domain')
	parser.add_argument('--config', dest='config', nargs='?',
						help='Manually specify config file for testing')
	
	args = parser.parse_args()

	sitesDir = os.path.expanduser("/var/www")
	projectFolder = os.path.join(sitesDir, args.domain)
	sourceFolder = os.path.join(projectFolder, 'source')
	destinationFolder = os.path.join(projectFolder, 'site')
	if args.config:
		configFile = os.path.join(sourceFolder, 'config', args.config)
	else:
		configFile = os.path.join(sourceFolder, 'config', args.domain)


	if args.subcommand == "build":
		pureStatic = PureStatic(configFile, sourceFolder, destinationFolder)
		pureStatic.build()
	elif args.subcommand == "watch":
		pureStatic = PureStatic(configFile, sourceFolder, destinationFolder)
		pureStatic.watch()
	elif args.subcommand == 'create':
		create_site(args.domain, isServer=args.server)
	elif args.subcommand == 'delete':
		delete_site(args.domain, isServer=args.server)
		

	# FINISH HERE: Do something with these values!
	# Use a config file to determine project folder location
	# Figure out multiple user setup. Need groups.
	# Who's running what as who and needs permissions to where? Git?
	# Need sudo for reloading nginx on server. Oh noes.

